#!/usr/bin/env python

"""
Radius user creation CGI

A basic user-creation device which inserts a user in a MySQL-backed
Radius server, with account creation rate throttling to prevent the
auth server from being flooded by malicious users.

First, a "human identifier" image is created and the value stored in
a local SQLite database to prevent replay attacks.

The content of the image is obfuscated with 3des and stored in the GET
string.  No session data is used.

mkershaw@arubanetworks.com
http://labs.arubanetworks.com

"""

# Do we store recently used images in a sqlite db to prevent replay?
USE_SQLITE_VERIFY = 1

import pyDes, cgi, cgitb, random, MySQLdb, time
from sys import exit
if USE_SQLITE_VERIFY:
    from pysqlite2 import dbapi2 as sqlite

# Shared private key, must match the key in image.cgi
PRIVATEKEY="010101010101010101010101"

# Path to a writeable sqlite database for storing recent
# attempts, with the schema:
# CREATE TABLE captcha (inserted int, captcha char(5));
PASSDB="/path/to/sqlite/passdb"

# MYSQL server used by Radius
MYSQLSRV="10.10.100.5"
# Username and password for a user able to select from db.usergroup
# and insert into db.usergroup and db.radcheck
MYSQLUSR="radiuscgi"
MYSQLPWD="radiuscgi"
# Database loaded with freeradius schema
MYSQLDB="radius"

# Group to add new users to
RADIUSGROUP="demo"

def printform(comment):
    myvals = { }

    myvals["comment"] = comment

    letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ23456789"
    word = ''.join([random.choice(letters) for i in range(5)])
    rw = ''.join([random.choice(letters) for i in range(random.randint(5, 20))])

    try:
        k = pyDes.triple_des(PRIVATEKEY)
    except:
        print "Content-type: text/html\n"
        print "form failure"
        exit(0)

    myvals["hash"] = (k.encrypt("%s:%s:%s" % (rw, time.time(), word), "*")).encode("hex")

    print "Content-type: text/html\n"
    print (open("form.phtml", "r").read() % myvals)

def printl2(user, pw):
    # time:random:user:pw

    letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ23456789"
    word = ''.join([random.choice(letters) for i in range(5)])
    rw = ''.join([random.choice(letters) for i in range(random.randint(5, 20))])

    try:
        k = pyDes.triple_des(PRIVATEKEY)
    except:
        print "Content-type: text/html\n"
        print "form failure"
        exit(0)

    hash = (k.encrypt("%s:%s:%s:%s" % (rw, time.time(), user.encode("hex"), pw.encode("hex")), "*")).encode("hex")

    print "Location: radius.cgi?l2h=%s\n" % hash

def checkcaptcha(captcha):
    try:
        con = sqlite.connect(PASSDB)
    except:
        printform("Captcha DB failure")
        exit(0)

    cursor = con.cursor();

    try:
        cursor.execute("SELECT * FROM captcha WHERE captcha = %s" % (MySQLdb.string_literal(captcha)))
    except:
        printform("Text match DB failure")
        exit(0)

    if cursor.rowcount:
        printform("Text already used")
        exit(0)

    try:
        cursor.execute("INSERT INTO captcha (inserted, captcha) VALUES (%d, %s)" % (int(time.time()), MySQLdb.string_literal(captcha)))
    except:
        printform("Text DB failure")
        exit(0)

    con.commit()

    try:
        # Clean up the captcha DB
        cursor.execute("DELETE FROM captcha WHERE inserted < %d" % int(time.time() - 30))
    except:
        return 1

    con.commit()

    return 1

def insert(user, pw):
    try:
        conn = MySQLdb.connect (host = MYSQLSRV,
                                user = MYSQLUSR,
                                passwd = MYSQLPWD,
                                db = MYSQLDB)
    except:
        printform("SQL server not available")

    cursor = conn.cursor()
    cursor.execute("SELECT UserName from usergroup WHERE UserName = %s;" % MySQLdb.string_literal(user))
    if cursor.rowcount:
        printform("Username already exists")
        exit(0)

    cursor.execute("INSERT INTO radcheck (username, attribute, value, op) values (%s, \"Cleartext-Password\", %s, \":=\");" % (MySQLdb.string_literal(user), MySQLdb.string_literal(pw)))
    if not cursor.rowcount:
        printform("Error")
        exit(0)

    cursor.execute("INSERT INTO usergroup (username, groupname) values (%s, \"%s\");" % (MySQLdb.string_literal(user), RADIUSGROUP))
    if not cursor.rowcount:
        printform("Error")
        exit(0)

    print "Content-type: text/html\n"
    print (open("done.phtml", "r").read())
    exit(0)

# Main processing
form = cgi.FieldStorage()

if not form:
    printform("")
    exit(0)

hash = form.getvalue("hash")
word = form.getvalue("captcha")
user = form.getvalue("username")
pw = form.getvalue("passwd")
pw2 = form.getvalue("passwd2")
l2h = form.getvalue("l2h")

try:
    k = pyDes.triple_des(PRIVATEKEY)
except:
    printform("Hash failure")
    exit(0)

# Second-stage hash actually triggers submit
if not l2h == None:
    try:
        s = k.decrypt(l2h.decode("hex"), "*")
    except:
        printform("L2 Hash failure")
        exit(0)

    v = s.split(":")
    # random:time:userhex:pwhex
    if len(v) != 4:
        printform("Invalid L2 hash")
        exit(0)

    try:
        ptime = float(v[1])
    except:
        printform("Invalid L2 hash")

    user = v[2].decode("hex")
    pw = v[3].decode("hex")

    # L2 hash only lives for 5 seconds
    if time.time() - ptime > 5:
        printform("L2 Hash expired")
        exit(0)

    insert(user, pw)
    print "exit"
    exit(0)

if (word == None or hash == None or user == None or pw == None or pw2 == None):
    printform("Missing form values")
    exit(0)

if len(pw) < 5:
    printform("Password should be at least 5 characters")
    exit(0)

if not pw == pw2:
    printform("Passwords didn't match")
    exit(0)

try:
    s = k.decrypt(hash.decode("hex"), "*")
except:
    printform("Hash failure")
    exit(0)

v = s.split(":")
if (len(v) != 3):
    printform("Invalid hash")
    exit(0)

if (v[2] == word.upper()):
    # Checkcaptcha will abort internally if it fails
    if USE_SQLITE_VERIFY:
        checkcaptcha(word.upper())
    printl2(user, pw)
    exit(0)

printform("invalid word")

